/*
  Copyright (C) 2006 Imperial College London and others.

  Please see the AUTHORS file in the main source directory for a full list
  of copyright holders.

  Gerard Gorman
  Applied Modelling and Computation Group
  Department of Earth Science and Engineering
  Imperial College London

  g.gorman@Imperial.ac.uk

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
  USA
*/

#include "Adaptivity.h"
#include "cinterfaces.h"

#ifdef HAVE_MPI
#include <mpi.h>
#endif

#include <cassert>
#include <cmath>
#include <iostream>
#include <deque>
#include <string>

#include <string.h>

using namespace std;

// Constructer
Adaptivity::Adaptivity(){
  MyRank = 0;
  NProcs = 1;
#ifdef HAVE_MPI
  if(MPI::Is_initialized()){
    MyRank = MPI::COMM_WORLD.Get_rank();
    NProcs = MPI::COMM_WORLD.Get_size();
  }
#endif
  
  newNNodes     = 0;
  newNElements  = 0;
  newNSElements = 0;

  newX = NULL;
  newY = NULL;
  newZ = NULL;

  numberingOffset = 1;
  
  volumeID = NULL;

  NPrivateNodes = 0;

  set_adapt_sweeps(10);

  AdaptOpts[1] = LOGICAL_TRUE; // collapse edges if true
  disableGeometryDiscovery();
  enableMovementConnectivity();
  enableRefinement();
  disableSurfaceLock();

  setFunctionalTolerance(0.7);

  nloc  = 4;
  snloc = 3;
  
  NWSZEN = 0;
  NWSZSN = 0;
  NWSZNN = 0;
  NWNDLC = 0;
  NWSROW = 0;
  NWENLB = 0;
  NWENLS = 0;
  NWSNLB = 0; 
  NWSNLS = 0; 
  NWSFID = 0;
  NWELRG = 0;
  NWNODX = 0; 
  NWNODY = 0;
  NWNODZ = 0;
  NEWMTX = 0;
  NEWFLD = 0;
  ADPBIG = 0; 
  ADPNOD = 0;

  InterpolationOn();
};

afloat_t Adaptivity::edgeLengthDistribution(afloat_t w){
#ifdef HAVE_VTK
  int edgecount = 0;
  int badedges = 0;
  const int dim = 3;
  
  // Figure out the edges.
  std::set<std::pair<int, int> > edges;

  // Form NNlist.
  for(int e=0;e<ug->GetNumberOfCells();e++){
    int k, l;
    vtkTetra *tetra = (vtkTetra *)ug->GetCell(e);
    for(size_t i=0;i<4;i++){
      for(size_t j=i+1;j<4;j++){
        k = tetra->GetPointId(i);
        l = tetra->GetPointId(j);
        edges.insert(std::pair<int, int>(min(k, l), max(k, l)));
      }
    }
  }
  
  for (std::set<std::pair<int, int> >::iterator edge = edges.begin(); edge != edges.end(); edge++){
    edgecount++;
    // First, compute the metric over the edge by averaging the nodal metrics.
    afloat_t edge_metric[dim*dim];
    for (int i = 0; i < dim; i++){
      for (int j = 0; j < dim; j++){
        edge_metric[i*dim+j] = (Metric[edge->first*dim*dim + i*dim + j] + Metric[edge->second*dim*dim + i*dim + j]) / 2.0;
      }
    }
    
    // Now, compute the inner product.
    afloat_t vec[dim];
    vec[0] = X[edge->first] - X[edge->second];
    vec[1] = Y[edge->first] - Y[edge->second];
    vec[2] = Z[edge->first] - Z[edge->second];

    afloat_t len = 0;
    for (int i = 0; i < dim; i++){
      afloat_t tmp = 0;
      for (int j = 0; j < dim; j++)
        tmp = tmp + edge_metric[i*dim + j] * vec[j];
      len = len + vec[i] * tmp;
    }
    len = sqrt(len);
    if ((len < 1.0 - w) || (len > 1.0 + w))
      badedges++;
  }
  
#ifdef HAVE_MPI
  int gbadedges, gedgecount;
  MPI::COMM_WORLD.Allreduce(&badedges, &gbadedges, 1, MPI::INT, MPI::SUM);
  MPI::COMM_WORLD.Allreduce(&edgecount, &gedgecount, 1, MPI::INT, MPI::SUM);
  return gbadedges / (afloat_t)gedgecount;
#else
  return badedges / (afloat_t)edgecount;
#endif

#else
  cerr<<"WARNING: VTK support has not been compiled.\n";
  return 0.0;
#endif
}

void Adaptivity::adapt(){
  // Only does 3D
  int Geom3D = LOGICAL_TRUE;
  
  // DOTOP is the minimum element functional to allow to be changed (min 0.1)
  //afloat_t DOTOP  = (fabs(MESTP1)>0.15)?fabs(MESTP1):0.15;
  afloat_t DOTOP = 0.5; 

  //  MINCHG is the minimum relative change allowed to a local functional (min 0.01)
  //afloat_t MINCHG = 0.01;
  afloat_t MINCHG = 0.3;

  // If true then use the quality of the element rather than insphere
  int USEQ = LOGICAL_FALSE;
    
  NWSZEN = 0;
  NWSZSN = 0;
  NWSZEN = 0;
  NWSZSN = 0;
  
  // These variables...if they are set to a non-negitive number they
  // write back the data - otherwise no
  NEWMTX =  1; // New metric values
  NWNDLC =  1; // Indicates which old element contains which new node
  NWSROW = -1;
  
  //
  // Local variables
  //
  
  // Find the expected number of elements given a metric.
  int NNodes = X.size();
  int NElements = ENList.size()/nloc;
  int NSElements = SENList.size()/snloc;
  int XPCTEL = max(get_predicted_nelements_fc(&(Metric[0]), &(X[0]), &(Y[0]), &(Z[0]),
                                              &(ENList[0]), &NNodes, &NElements, &nloc), max(NElements, 100000));

  // Calculate how much memory adaptivity needs and allocate it.
  int intBuffer_size, floatBuffer_size;
  int DUM1=-1, DUM2=-1,DUM3=-1;
  int flag=LOGICAL_TRUE;

  int totfre=0;
  for(vector<int>::const_iterator it=nfreedom.begin();it!=nfreedom.end();++it){
    totfre+=*it;
  }

  adaptmem_(&NNodes, &NElements, &DUM1, &NSElements, &totfre,
	    &XPCTEL, &DUM2, &DUM3, &flag, &intBuffer_size, &floatBuffer_size);

  intBuffer.resize(intBuffer_size);
  floatBuffer.resize(floatBuffer_size);
  
  vector<int> ENLBasePtr(NElements+1);
  for(int i=0; i<=NElements; i++)
    ENLBasePtr[i] = i*nloc;
  int sizeENList = ENList.size();
  
  vector<int> SENLBasePtr(NSElements+1);
  for(int i=0; i<=NSElements; i++)
    SENLBasePtr[i]=i*snloc;
  int sizeSENList = SENList.size();
  
  vector<int> ElementRegion;
  const int *_volumeID = volumeID;
  if(volumeID==NULL){
    ElementRegion.resize(NElements);
    _volumeID=&(ElementRegion[0]);
    for(int i=0; i<NElements; i++)
      ElementRegion[i] = 1;
  }
  
  if(surfID.size()==0){
    // Probably should calculate this using the co-planar patch
    // algorithm if it's not already done.
    surfID.resize(NSElements);
    for(int i=0;i<NSElements;i++)
      surfID[i]=1; 
  }
  
  // Serialise halo information.
  vector<int> Gather, ATOSEN(NProcs+1, 0), Scatter, ATOREC(NProcs+1, 0);
  if(NProcs>1){
    ATOSEN[0] = 1;
    ATOREC[0] = 1;
    for(int p=0;p<NProcs;p++){
      ATOSEN[p+1] = ATOSEN[p]+send_nodes[p].size();
      for(set<int>::iterator it=send_nodes[p].begin();it!=send_nodes[p].end();++it){
        Gather.push_back(gnn2lnn[*it]+1);
      }
      ATOREC[p+1] = ATOREC[p]+recv_nodes[p].size();
      for(set<int>::iterator it=recv_nodes[p].begin();it!=recv_nodes[p].end();++it){
        Scatter.push_back(gnn2lnn[*it]+1);
      }
    }
  }
  int NGather = Gather.size();
  int NScatter = Scatter.size();

  int debug_level = 0;
  int *PRDNDS = NULL, NPRDND=0;

  chcnsy = LOGICAL_FALSE;
  dbg    = LOGICAL_FALSE;
    
  // -1 lets libadapt figure it out for itself what the maximum number
  // -of nodes is.
  int mxnods = -1;

  // Call the main adaptivity routine  
  adptvy_(&(intBuffer[0]), &intBuffer_size, &(floatBuffer[0]), &floatBuffer_size,
	  &Geom3D, &SRFGMY, &USEQ,
	  &NNodes, &NElements, &NSElements, &mxnods,
	  &sizeENList, &(ENLBasePtr[0]), &(ENList[0]), _volumeID,
	  &CLCGMY, 
	  &sizeSENList, &(SENLBasePtr[0]), &(SENList[0]), &(surfID[0]),
	  PRDNDS, &NPRDND,
	  &(X[0]), &(Y[0]), &(Z[0]),
	  &NNodes, &NElements, &sizeENList, &(ENList[0]), &(ENLBasePtr[0]),
	  &(X[0]), &(Y[0]), &(Z[0]),
	  &(Metric[0]), &(fields[0]), &(nfreedom[0]), &totfre, &nfields,
	  &XPCTEL,
	  &newNNodes, &newNElements, &newNSElements,
	  &NWSZEN, &NWSZSN, &NWSZNN, &NWNDLC, &NWSROW,
	  &NWENLB, &NWENLS, &NWSNLB, &NWSNLS, &NWSFID,
	  &NWELRG, &NWNODX, &NWNODY, &NWNODZ,
	  &NEWMTX, &NEWFLD,
	  &ADPBIG, &ADPNOD,
	  &DOTOP,  &MINCHG, &MaxNumberAdaptIterations, AdaptOpts, &TWOSTG, &TOGTHR,
	  &(Gather[0]), &(Scatter[0]), &NGather, &NScatter, &NPrivateNodes,
	  &(ATOSEN[0]), &(ATOREC[0]),  &NProcs, &debug_level, &dbg, &chcnsy);

  // Zero tolerence for errors from adaptivity
  assert(newNNodes>0);
  NPrivateNodes = newNNodes-NScatter;

  newX = &(floatBuffer[NWNODX-1]);
  newY = &(floatBuffer[NWNODY-1]);
  newZ = &(floatBuffer[NWNODZ-1]);

  // Get new global number offsets.
  vector<int> offsets(NProcs+1, 0);
#ifdef HAVE_MPI
  if(NProcs>1)
    MPI::COMM_WORLD.Allgather(&NPrivateNodes, 1, MPI::INT, &(offsets[1]), 1, MPI::INT);
#endif
  for(int i=1;i<=NProcs;i++)
    offsets[i] += offsets[i-1];

  if(NProcs==1)
    NPrivateNodes = newNNodes;

  // Reconstruct halo and global-local node numbering lut.
  gnn2lnn.clear();
  lnn2gnn.clear();
  vector<int> local_lnn2gnn(newNNodes, -1);

  send_nodes.clear();
  recv_nodes.clear();
  
  for(int i=0; i<NPrivateNodes;i++)
    local_lnn2gnn[i] = offsets[MyRank]+i;
  
  if(NProcs>1){
    vector< vector<int> > send_nodes_vec(NProcs);
    for(int p=0;p<NProcs;p++){
      for(int i=(ATOSEN[p]-1);i<(ATOSEN[p+1]-1);++i){
        int lnn = Gather[i]-1;
        assert(lnn<NPrivateNodes);
        assert(local_lnn2gnn[lnn]!=-1);
        int gnn = local_lnn2gnn[lnn];
        send_nodes_vec[p].push_back(gnn);
      }
    }
    
    vector<int> recv_size(NProcs), send_size(NProcs);
    for(int p=0;p<NProcs;p++)
      send_size[p] = send_nodes_vec[p].size();

    MPI::COMM_WORLD.Alltoall(&(send_size[0]), 1, MPI::INT,
                             &(recv_size[0]), 1, MPI::INT);

    vector<MPI::Request> sendRequest(NProcs, MPI::REQUEST_NULL);
    vector<MPI::Request> recvRequest(NProcs, MPI::REQUEST_NULL);
    MPI::Status recvStatus;
    
    for(int p=0;p<NProcs;p++){
      if(send_size[p]){
        sendRequest[p] = MPI::COMM_WORLD.Isend(&(send_nodes_vec[p][0]), send_size[p], MPI::INT, p, 1);
      }
    }
    
    vector< vector<int> > recv_nodes_vec(NProcs);
    for(int p=0;p<NProcs;p++){
      if(recv_size[p]){
        recv_nodes_vec[p].resize(recv_size[p]);
        recvRequest[p] = MPI::COMM_WORLD.Irecv(&(recv_nodes_vec[p][0]), recv_size[p], MPI::INT, p, 1);
      }
    }
    
    MPI::Request::Waitall(NProcs, &(sendRequest[0])); // Wait all sends
    MPI::Request::Waitall(NProcs, &(recvRequest[0])); // Wait all receives

    //
    recv_nodes.resize(NProcs);
    send_nodes.resize(NProcs);
    for(int p=0;p<NProcs;p++){
      for(int i=0;i<recv_size[p];++i){
        int gnn = recv_nodes_vec[p][i];
        int lnn = Scatter[ATOREC[p]-1+i]-1;
        assert(lnn>=NPrivateNodes);

        recv_nodes[p].insert(gnn);
        assert(local_lnn2gnn[lnn]=-1);
        local_lnn2gnn[lnn] = gnn;
      }
      for(vector<int>::iterator it=send_nodes_vec[p].begin();it!=send_nodes_vec[p].end();++it)
        send_nodes[p].insert(*it);
    }
  }
  lnn2gnn.swap(local_lnn2gnn);

  for(int i=0;i<newNNodes;i++){
    assert(lnn2gnn[i]!=-1);
    gnn2lnn[lnn2gnn[i]] = i;
  }
}

void Adaptivity::enableGeometryDiscovery(){
  CLCGMY = LOGICAL_TRUE;
}

void Adaptivity::enableMovementConnectivity(){
  TOGTHR = LOGICAL_TRUE;
}

void Adaptivity::enableRefinement(){
  TWOSTG = LOGICAL_FALSE;

  AdaptOpts[0] = LOGICAL_TRUE;
  AdaptOpts[2] = LOGICAL_TRUE;
  AdaptOpts[3] = LOGICAL_TRUE;
  AdaptOpts[4] = LOGICAL_TRUE;
  AdaptOpts[5] = LOGICAL_TRUE;
}

void Adaptivity::enableSurfaceLock(){
  SRFGMY = LOGICAL_TRUE;
}

void Adaptivity::disableGeometryDiscovery(){
  CLCGMY = LOGICAL_FALSE;
}

void Adaptivity::disableMovementConnectivity(){
  TOGTHR = LOGICAL_FALSE;
}

void Adaptivity::disableRefinement(){
  TWOSTG = LOGICAL_TRUE;

  AdaptOpts[0] = LOGICAL_FALSE;
  AdaptOpts[2] = LOGICAL_FALSE;
  AdaptOpts[3] = LOGICAL_FALSE;
  AdaptOpts[4] = LOGICAL_FALSE;
  AdaptOpts[5] = LOGICAL_FALSE;
}

void Adaptivity::disableSurfaceLock(){
  SRFGMY = LOGICAL_FALSE;
}

#ifdef HAVE_VTK
/// Get a vtkUnstructuredGrid object that contains the adapted mesh.
vtkUnstructuredGrid* Adaptivity::get_adapted_vtu(int &_NPrivateNodes){
  _NPrivateNodes = NPrivateNodes;
  return get_adapted_vtu();
}

vtkUnstructuredGrid* Adaptivity::get_adapted_vtu(){
  vtkUnstructuredGrid *aug = vtkUnstructuredGrid::New();

  // Set up points.
  vtkPoints *points = vtkPoints::New();
  points->SetDataTypeToDouble();
  points->SetNumberOfPoints(newNNodes);
  for(int i=0;i<newNNodes;++i){
    points->SetPoint(i, newX[i], newY[i], newZ[i]);
  }
  aug->SetPoints(points);
  aug->Update();
  points->Delete();
  
  // Add GlobalNodeNumbers.
  vtkIntArray *GlobalNodeNumbers = vtkIntArray::New();
  GlobalNodeNumbers->SetName("GlobalNodeNumbers");
  GlobalNodeNumbers->SetNumberOfTuples(newNNodes);
  GlobalNodeNumbers->SetNumberOfComponents(1);
  for(int i=0;i<newNNodes;i++){
    GlobalNodeNumbers->SetTuple1(i, lnn2gnn[i]);
  }
  aug->GetPointData()->AddArray(GlobalNodeNumbers);
  aug->Update();
  GlobalNodeNumbers->Delete();

  // Set up elements.
  int *newENList = &(intBuffer[NWENLS-1]);
  for(int i=0;i<newNElements;i++){
    vtkIdType pts[4];
    //for(size_t j=0;i<4;j++)
    for(int j=3;j>=0;j--)
      pts[j] = newENList[i*4+j] - 1;
    aug->InsertNextCell(VTK_TETRA, 4, pts);
  }
  
  if(NProcs>1){
    vtkIntArray *nowner_vtk = vtkIntArray::New();
    nowner_vtk->SetNumberOfComponents(1);
    nowner_vtk->SetNumberOfTuples(newNNodes);
    nowner_vtk->SetName("NodeOwner");
    for(int i=0;i<newNNodes;i++)
      nowner_vtk->SetTuple1(i, MyRank);

    vector<int> nowner(newNNodes, MyRank);
    for(int p=0;p<NProcs;p++){
      for(set<int>::iterator it=recv_nodes[p].begin();it!=recv_nodes[p].end();++it){
        assert(gnn2lnn.find(*it)!=gnn2lnn.end());
        nowner[gnn2lnn[*it]] = p;
        nowner_vtk->SetTuple1(gnn2lnn[*it], p);
      }
    }
    aug->GetPointData()->AddArray(nowner_vtk);
    aug->Update();
    nowner_vtk->Delete();

    vtkIntArray *eowner = vtkIntArray::New();
    eowner->SetNumberOfComponents(1);
    eowner->SetNumberOfTuples(newNElements);
    eowner->SetName("ElementOwner");
    for(int e=0;e<newNElements;e++){
      int owner=MyRank;
      for(int i=0;i<4;i++)
        owner = min(owner, nowner[newENList[e*4+i] - 1]);
      eowner->SetTuple1(e, owner);
    }
    aug->GetCellData()->AddArray(eowner);
    aug->Update();
    eowner->Delete();
  }

  // Copy back metric
  vtkDoubleArray *m = vtkDoubleArray::New();
  afloat_t *newMetric = &(floatBuffer[NEWMTX-1]);
  m->SetNumberOfComponents(9);
  m->SetNumberOfTuples(newNNodes);
  m->SetName("metric");
  for(int i=0;i<newNNodes;i++){
    m->SetTuple9(i,
                 newMetric[i*9  ], newMetric[i*9+1], newMetric[i*9+2],
                 newMetric[i*9+3], newMetric[i*9+4], newMetric[i*9+5],
                 newMetric[i*9+6], newMetric[i*9+7], newMetric[i*9+8]);
  }
  aug->GetPointData()->AddArray(m);
  aug->Update();
  m->Delete();

  if(interpolate){
    // Copy interpolated fields.
    afloat_t *new_fields = &(floatBuffer[NEWFLD-1]);
    size_t pos=0;
    int total_nfields = ug->GetPointData()->GetNumberOfArrays();
    for(int n=0;n<total_nfields;n++){
      vtkDataArray *array = ug->GetPointData()->GetArray(n);
      if(string("metric")==string(array->GetName())||
         string("GlobalNodeNumbers")==string(array->GetName()))
	continue;
      
      size_t ncomp = array->GetNumberOfComponents();
      
      vtkDoubleArray *m = vtkDoubleArray::New();
      m->SetNumberOfComponents(ncomp);
      m->SetNumberOfTuples(newNNodes);
      m->SetName(array->GetName());

      vector<double> tuple(ncomp);
      for(int i=0;i<newNNodes;i++){
	for(size_t j=0;j<ncomp;j++)
	  tuple[j] = new_fields[pos++];
	m->SetTuple(i, &(tuple[0]));
      }
      aug->GetPointData()->AddArray(m);
      aug->Update();
      m->Delete();
    }
  }
  
  return aug;
}
#endif

void Adaptivity::getMeshDimensions(int *_NNodes, int *_NElements, int *_NSElements){
  *_NNodes     = newNNodes;
  *_NElements  = newNElements;
  *_NSElements = newNSElements;
}

void Adaptivity::get_surface_ids(vector<int> &sids){
  sids.resize(newNSElements);
  for(int i=0;i<newNSElements;i++)
    sids[i] = intBuffer[NWSFID-1+i];
}

void Adaptivity::get_surface_mesh(vector<int> &SENList){
  SENList.resize(3*newNSElements);
  for(int i=0;i<newNSElements*3;i++)
    SENList[i] = intBuffer[NWSNLS-1+i]-1;
}

void Adaptivity::getVertices(afloat_t *_X, afloat_t *_Y, afloat_t *_Z){
  memcpy(_X, &(floatBuffer[NWNODX-1]), newNNodes*sizeof(afloat_t));
  memcpy(_Y, &(floatBuffer[NWNODY-1]), newNNodes*sizeof(afloat_t));
  memcpy(_Z, &(floatBuffer[NWNODZ-1]), newNNodes*sizeof(afloat_t));
}

void Adaptivity::getVolumeID(int *_newvolumeID){
  memcpy(_newvolumeID, &(intBuffer[NWELRG-1]), newNElements*sizeof(int));
}

void Adaptivity::getVolumeMesh(int *_newENList){
  memcpy(_newENList, &(intBuffer[NWENLS-1]), 4*newNElements*sizeof(int));
}

void Adaptivity::InterpolationOn(){
  interpolate = true;
}

void Adaptivity::InterpolationOff(){
  interpolate = false;
}

void Adaptivity::setFunctionalTolerance(afloat_t tol){
  MESTP1 = tol;
}

void Adaptivity::set_adapt_sweeps(int _iterations){
  MaxNumberAdaptIterations = _iterations;
}

void Adaptivity::set_metric(afloat_t *_metric){
  size_t NNodes = X.size();
  Metric.resize(9*NNodes);
  for(size_t i=0;i<9*NNodes;i++){
    Metric[i] = _metric[i];
  }
}

void Adaptivity::set_surface_ids(vector<int> &_sids){
  if(SENList.empty()){
    cerr<<"WARNING: trying to set surface id's but no surface has been specified\n";
    return;
  }
  
  surfID = _sids;
  
  return;
}

void Adaptivity::set_surface_mesh(vector<int> &_SENList){
  SENList = _SENList;
  for(vector<int>::iterator it=SENList.begin();it!=SENList.end();++it)
    *it = *it + 1;

  return;
}

void Adaptivity::set_points(afloat_t *_X, afloat_t *_Y, afloat_t *_Z, int _NNodes){
  NPrivateNodes = _NNodes;
  X.resize(_NNodes);
  Y.resize(_NNodes);
  Z.resize(_NNodes);
  for(int i=0;i<_NNodes;i++){
    X[i]=_X[i];
    Y[i]=_Y[i];
    Z[i]=_Z[i];
  }

  return;
}

void Adaptivity::setVolumeID(int *_volumeID){
  volumeID = _volumeID;
}

void Adaptivity::set_volume_mesh(int *_ENList, int _NElements){
  ENList.resize(4*_NElements);
  for(int i=0;i<4*_NElements;i++)
    ENList[i] = _ENList[i];

  return;
}

#ifdef HAVE_VTK
/// Set the input from a vtkUnstructuredGrid
void Adaptivity::SetInput(vtkUnstructuredGrid *_ug){
  int _NPrivateNodes = _ug->GetNumberOfPoints();
  SetInput(_ug, _NPrivateNodes);
}

void Adaptivity::SetInput(vtkUnstructuredGrid *_ug, int _NPrivateNodes){
  ug = _ug;
  int NNodes = ug->GetNumberOfPoints();
  NPrivateNodes = _NPrivateNodes;
  
#ifdef HAVE_MPI
  // Construct receive part of halo and construct global-local node
  // numbering lut.
  vector<int> offsets(NProcs+1, 0);
  if(NProcs>1){
    // Calculate offsets
    MPI::COMM_WORLD.Allgather(&NPrivateNodes, 1, MPI::INT, &(offsets[1]), 1, MPI::INT);
    for(int i=1;i<=NProcs;i++)
      offsets[i] += offsets[i-1];
    
    // Create lnn2gnn and gnn2lnn.
    lnn2gnn.resize(NNodes);
    for(int i=0; i<NNodes;i++){
      int gnn = ug->GetPointData()->GetArray("GlobalNodeNumbers")->GetTuple1(i);
      gnn2lnn[gnn] = i;
      lnn2gnn[i] = gnn;
    }
    
    // Create receive halo.
    recv_nodes.clear();
    recv_nodes.resize(NProcs);
    for(int i=NPrivateNodes;i<NNodes;i++){
      int gnn = lnn2gnn[i];
      
      int owner = -1;
      for(int p=0;p<NProcs;p++){
        if(gnn<offsets[p+1]){
          owner = p;
          break;
        }
      }
      assert(owner!=-1);
      assert(recv_nodes[owner].count(gnn)==0);
      recv_nodes[owner].insert(gnn);
    }
    
    // Construct send part of halo.
    send_nodes.clear();
    send_nodes.resize(NProcs);
    vector<int> recv_size(NProcs), send_size(NProcs);
    for(int p=0;p<NProcs;p++)
      recv_size[p] = recv_nodes[p].size();
    MPI::COMM_WORLD.Alltoall(&(recv_size[0]), 1, MPI::INT,
                             &(send_size[0]), 1, MPI::INT);
    
    vector<MPI::Request> sendRequest(NProcs, MPI::REQUEST_NULL);
    vector<MPI::Request> recvRequest(NProcs, MPI::REQUEST_NULL);
    MPI::Status recvStatus;
    
    // Send
    vector< vector<int> > recv_nodes_vec(NProcs);
    for(int p=0;p<NProcs;p++){
      for(set<int>::const_iterator it=recv_nodes[p].begin(); it!=recv_nodes[p].end(); ++it){
        recv_nodes_vec[p].push_back(*it);
      }
      if(recv_size[p]){
        sendRequest[p] = MPI::COMM_WORLD.Isend(&(recv_nodes_vec[p][0]), recv_size[p], MPI::INT, p, 1);
      }
    }
    
    // Receive
    vector< vector<int> > send_nodes_vec(NProcs);
    for(int p=0;p<NProcs;p++){
      if(send_size[p]){
        send_nodes_vec[p].resize(send_size[p]);
        recvRequest[p] = MPI::COMM_WORLD.Irecv(&(send_nodes_vec[p][0]), send_size[p], MPI::INT, p, 1);
      }
    }
    
    MPI::Request::Waitall(NProcs, &(sendRequest[0])); // Wait all sends
    MPI::Request::Waitall(NProcs, &(recvRequest[0])); // Wait all receives
    
    for(int p=0;p<NProcs;p++){
      for(vector<int>::const_iterator it=send_nodes_vec[p].begin(); it!=send_nodes_vec[p].end(); ++it){
        send_nodes[p].insert(*it);
      }
    }
  }
#endif

  X.resize(NNodes);
  Y.resize(NNodes);
  Z.resize(NNodes);
  double r[3];
  for(int i=0;i<NNodes;i++){
    ug->GetPoints()->GetPoint(i, r);
    X[i] = r[0];
    Y[i] = r[1];
    Z[i] = r[2];
  }
  
  int NElements = ug->GetNumberOfCells();
  ENList.resize(4*NElements);
  for(int i=0;i<NElements;i++){
    assert(ug->GetCell(i)->GetCellType()==VTK_TETRA);
    vtkTetra *tetra = (vtkTetra *)ug->GetCell(i);
    
    for(int j=3;j>=0;j--){
      ENList[i*4+j] = tetra->GetPointId(j) + 1;
    }
  }

  vtkDataArray *m = ug->GetPointData()->GetArray("metric");
  if(m!=NULL){
    Metric.resize(9*NNodes);
#ifdef DOUBLEP
    for(size_t i=0;i<NNodes;i++){
      m->GetTuple(i, &(Metric[i*9]));      
    }
#else
    double lMetric[9];
    for(int i=0;i<NNodes;i++){
      m->GetTuple(i, lMetric);
      for(size_t j=0;j<9;j++){
        Metric[9*i+j]=lMetric[j];
      }
    }
#endif
  }

  if(interpolate){
    nfields = ug->GetPointData()->GetNumberOfArrays();
    for(int i=0;i<ug->GetPointData()->GetNumberOfArrays();i++){
      vtkDataArray *array = ug->GetPointData()->GetArray(i);
      if(string("metric")==string(array->GetName())||
         string("GlobalNodeNumbers")==string(array->GetName())){
        nfields--;
	continue;
      }

      size_t ncomp = array->GetNumberOfComponents();
      size_t ntuples = array->GetNumberOfTuples();
      nfreedom.push_back(ncomp);
      for(size_t t=0;t<ntuples;t++){
	double tuple[9];
	array->GetTuple(t, tuple);
	for(size_t j=0;j<ncomp;j++)
	  fields.push_back(tuple[j]);
      }
    }
  }

  return;
}
#endif

double Adaptivity::volume(int n1, int n2, int n3, int n4) const{
  n1 = n1 - numberingOffset;
  n2 = n2 - numberingOffset;
  n3 = n3 - numberingOffset;
  n4 = n4 - numberingOffset;
  
  return volume(X[n1], Y[n1], Z[n1],
		X[n2], Y[n2], Z[n2],
		X[n3], Y[n3], Z[n3],
		X[n4], Y[n4], Z[n4]);
}

double Adaptivity::volume(int n1, int n2, int n3, afloat_t x, afloat_t y, afloat_t z) const{
  n1 = n1 - numberingOffset;
  n2 = n2 - numberingOffset;
  n3 = n3 - numberingOffset;
  
  return volume(X[n1], Y[n1], Z[n1],
		X[n2], Y[n2], Z[n2],
		X[n3], Y[n3], Z[n3],
		x, y, z);
}


double Adaptivity::volume(afloat_t x1, afloat_t y1, afloat_t z1,
                          afloat_t x2, afloat_t y2, afloat_t z2,
                          afloat_t x3, afloat_t y3, afloat_t z3,
                          afloat_t x4, afloat_t y4, afloat_t z4) const{
  double vx1 = x2 - x1;
  double vy1 = y2 - y1;
  double vz1 = z2 - z1;
  
  double vx2 = x3 - x1;
  double vy2 = y3 - y1;
  double vz2 = z3 - z1;
  
  double vx3 = x4 - x1;
  double vy3 = y4 - y1;
  double vz3 = z4 - z1;
  
  double det = 
    vx1*(vy2*vz3 - vy3*vz2) - 
    vx2*(vy1*vz3 - vy3*vz1) + 
    vx3*(vy1*vz2 - vy2*vz1);

  return det/6.0;
}
