/* Copyright (C) 2006 Imperial College London and others.

 Please see the AUTHORS file in the main source directory for a full list
 of copyright holders.

 Dr Gerard J Gorman
 Applied Modelling and Computation Group
 Department of Earth Science and Engineering
 Imperial College London

 g.gorman@imperial.ac.uk

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 USA
*/
// Function to convert the mesh from FLUIDITY's matrix format
// to horizons internal format.

#include "samtypes.h"
#include "Mesh.h"
#include "c++debug.h"

#include <string.h>

#include <cassert>
#include <cmath>
#include <deque>
#include <set>
#include <vector>

using namespace std;

// Limit the scope of these two sets to this file
static set<int> deleted_vol_elems;
static set<int> deleted_sur_elems;


void Mesh::import_fluidity(const int dim, const int NNOD, const int NELM, const int NSElements, const int stride,
			   const int field_len,
			   const samfloat_t X[],
			   const samfloat_t Y[],
			   const samfloat_t Z[],
			   const samfloat_t metric[],
			   const samfloat_t field[],
			   const int ENLIST[], const int nloc,
			   const int SNLIST[], const int SURFID[], const int snloc,
			   const int ATOREC[], const int SCATER[],
			   const int ATOSEN[], const int GATHER[]){
  
  deleted_vol_elems.clear();
  deleted_sur_elems.clear();

  dimension = dim;

  // Map nodal information to internal format.
  ECHO("Reading nodes into internal format... ");
  size_t flen = field_len;
  for(size_t n=0; n<(unsigned)NNOD; n++){
    SAM_Node node;
    
    node.set_gnn( n );
    node.set_current_owner( MyRank );
    
    // Nodewise field values.
    for(size_t f = 0; f<flen; f++)
      node.append_field(field + stride*f + n, 1);
    
    // Position.
    switch(dim){
      case 3:
        node.set_coord(X[n], Y[n], Z[n]);
	break;
      case 2:
        node.set_coord(X[n], Y[n]);
	break;
      default:
        ERROR("Invalid dimension");
    }
    
    // Nodewise metric used for adaptivity. This is handled seperately
    // from fields as it can be used to control the
    // graph-partitioning.
    node.set_metric(metric + n*dim*dim, dim*dim);
    
    CHECK(node);
    add_node(node);
  }
  ECHO("Read.");

  //
  // Correct owner information for halo nodes.
  // (j = ATOREC[i]-1) is a base pointer to a list 
  // of nodes to be received from partition i 
  // SCATTER[j] - 1 is the node in question.
  //
  vector< vector<unsigned> > halo( NProcs ); // This is handy later
  for(unsigned p=0; p<(unsigned)NProcs; p++){
    
    for(int j=(ATOREC[p]-1);j<(ATOREC[p+1]-1);j++){
      unsigned haloNode = SCATER[j] - 1;

      assert(j<ATOREC[NProcs]-1);          // Crap out!
      assert(haloNode<(unsigned)NNOD);     // Crap out!
      
      node_list[ haloNode ].set_current_owner( p );
      node_list[ haloNode ].set_flags( NODE_HALO ) ;
      halo[p].push_back( haloNode );
    }
  }
  do_node_headcount();
  
  // Establish UNNing for native nodes
  {
    // Set up universal node numbering offset.
    find_nodes_per_rank( num_nodes("native") );
    
    unn_offsets[0] = 0;
    for(int i = 0; i<NProcs; i++)
      unn_offsets[i+1] = unn_offsets[i] + nodes_per_rank[i];
    
    unsigned cnt=0;
    for(unsigned i=0; i<(unsigned)NNOD; i++){
      if(node_list[i].get_current_owner() == MyRank){
	node_list[i].set_unn( unn_offsets[MyRank]+cnt );
	cnt++;
      }
    }
  }
  
  //
  // Record the nodes that are shared.
  // (j = ATOSEN[i]-1) is a base pointer to a list 
  // of nodes to be sent to partition i 
  // GATHER[j] - 1 is the node in question.
  //
  ECHO( "Recording shared nodes...." );

  // shared_unns is created to maintain the order of the shared
  // nodes so that the respective halo_node unn's can be correctly 
  // updated.
  vector< vector<unsigned> > shared_unns(NProcs);
  for(unsigned p=0; p<(unsigned)NProcs; p++){
    for(int j=(ATOSEN[p]-1);j<(ATOSEN[p+1]-1);j++){
      unsigned haloNode = GATHER[j] - 1;
      unsigned unn = node_list[haloNode].get_unn();
      
      // Store such that the halo is always in order of increasing unn.
      shared_nodes[p].insert(unn);
      shared_unns[p].push_back( unn );
    }
  }
  ECHO( "Recorded." );

  ECHO("At this point the node data should be converted to the internal format.");
  ECHO("Now it is time to correct the UNN's for the halo nodes.");
  
  //
  // Correct the UNN of the halo nodes (aka. SCATER nodes)
  //
  {
    vector<unsigned> halo_cnt( NProcs );
    for(int i = 0; i<NProcs; i++)
      halo_cnt[i] = halo[i].size();
    
    vector< vector<unsigned> > halo_unns;
    halo_update(shared_unns, halo_cnt, halo_unns);

    // Write in the UNN's and halo_nodes map
    for(unsigned i=0; i<(unsigned)NProcs; i++){

      assert( halo[i].size() == halo_unns[i].size() );
      
      const unsigned len = halo_unns[i].size();
            
      for(unsigned j = 0; j<len; j++){
	unsigned haloNode = halo[i][j];

	ECHO( i << ":" << haloNode << " --> " << halo_unns[i][j] );

	node_list[ haloNode ].set_unn( halo_unns[i][j] );
	halo_nodes[i].insert(halo_unns[i][j]);
      }
    }
    
  }

  ECHO( "gnn" <<"\t" << "unn" << "\t" << "owner" << "\t" << "unn2gnn");
  for(deque<SAM_Node>::iterator it=node_list.begin(); it!=node_list.end(); ++it){
    CHECK( (*it).get_gnn() );
    CHECK( (*it).get_current_owner() );
    CHECK( unn2gnn( (*it).get_unn() ) );
  }

  ECHO("UNN's have been established. ");
  ECHO("Lets check UNN's. Printing the unn's of the nodes in gather");
  
  for(unsigned i=0; i<(unsigned)NProcs; i++){
    ECHO("Rank " << i << ":\t");
    
    ECHO("SCATER");
    for(int j = (ATOREC[i]-1); j< (ATOREC[i+1]-1); j++)
      ECHO("\t\t(gnn="<<SCATER[j]-1 <<") "<< node_list[SCATER[j]-1].get_unn());
    ECHO("GATHER");
    for(int j = (ATOSEN[i]-1); j< (ATOSEN[i+1]-1); j++)
      ECHO("\t\t(gnn="<< GATHER[j]-1 <<")"<< node_list[GATHER[j]-1].get_unn());
  }
  
  unsigned eid=0;
  { // convert volume element information
    ECHO( "Write volume elements to internal format. " );
    CHECK(NELM);
    CHECK(nloc);

    SAM_Element elem;
    for(int e=0; e<NELM; e++){
      CHECK(e);
      elem.set_flags(ELM_DEFAULT|ELM_VOLUME);
      bool wanted = false; // Maybe it's not even this domain!
      
      { // write element node list with UNN's
	int cnt = nloc;
	int pos = e*nloc;                                // Note: no -1 here
	vector<unn_t> nl(cnt);
	for(int n=0; n<cnt; n++){
	  int nodeNum = ENLIST[pos + n] - 1;
	  assert(nodeNum>=0   ); // or crap out
	  assert(nodeNum<NNOD ); // or crap out
	  nl[ n ] = node_list[nodeNum].get_unn();
          if( node_list[nodeNum].get_current_owner() == MyRank )
            wanted = true;
	}
	elem.set_enlist( nl );
      }
      
      if( !wanted ){
	deleted_vol_elems.insert(e);
	continue;
      }

      // USING_DISCONTINUOUS
      // Field values??

      elem.set_eid(eid++);

      ECHO("Adding volume element:");
      CHECK(elem);
      
      add_element( elem );
    }
  }
  
  { // Import surface.
    ECHO( "Import surface." );
    
    SAM_Element elem;
    for(int e=0; e<NSElements; e++){
      elem.set_flags(ELM_DEFAULT|ELM_SURFACE);
      bool wanted = false; // Maybe it's on the halo interface

      { // write element node list with UNN's
	int cnt = snloc;
	int pos = e*snloc;                                // Note: no -1 here
	vector<unn_t> nl(cnt);
	for(int n=0; n<cnt; n++){
	  int nodeNum = SNLIST[pos + n] - 1;
	  assert(nodeNum>=0   );
	  assert(nodeNum<NNOD );
	  nl[ n ] = node_list[nodeNum].get_unn();
          if( node_list[nodeNum].get_current_owner() == MyRank )
            wanted = true;
	}
        if(!wanted){
	  deleted_sur_elems.insert(e);
          continue;
	}

	elem.set_enlist( nl );
      }
      
      elem.set_eid(eid++);

      // add surface id's
      vector<int> ifield(1);
      ifield[0] = SURFID[e];
      elem.set_ifields( ifield );

      // USING_DISCONTINUOUS
      // Field values??
      ECHO( "Adding element:");
      CHECK( elem );

      add_element( elem );
    }
  do_element_headcount();
  }
 
  ECHO( "Finished translating fluidity stuff to internal format.");

  return;
}

void Mesh::import_pressure(const int NNOD, const int NELM,
			   const samfloat_t pressure[], const int nfields,
			   const int ENLIST[], const int ENLBAS[],
			   const int ATOREC[], const int SCATER[],
			   const int ATOSEN[], const int GATHER[]){
  // Make space for the pressure node list
  MFnode_list.clear();
  
  if(NNOD==0){
    
    // So the pressure mesh is the same as the velocity mesh so we can
    // just recreate it after we're finished.
    return;
    
  }else{
    
    // Set up UNNing for pressure nodes. As the number of pressure
    // nodes is always less or equal to the number of velocity nodes, and
    // the pressure node unn need not be contineous, the same offsets
    // will be used.    
    unsigned offset = 0;
    for(int i = 0; i<MyRank; i++)
      offset += nodes_per_rank[i];
    
    ECHO( "Reading pressure nodes into internal format... " );

    { // Map nodal information to internal format.
      for(unsigned n=0; n<(unsigned)NNOD; n++){
	PressureNode node;
	
	node.set_gnn( n );
	node.set_unn( offset + n );
	
	if(nfields>0){
	  vector<samfloat_t> flds(nfields);
	  memcpy(&(flds[0]), pressure+n*nfields, nfields*sizeof(samfloat_t));
	  node.set_pressure( flds );
	}
	
	MFnode_list.push_back( node );
      }
    }
    ECHO( "Read." );
    
    // shared_unns is created to maintain the order of the shared
    // nodes so that the respective halo_node unn's can be correctly 
    // updated.
    vector< vector<unsigned> > shared_unns(NProcs);
    for(unsigned p=0; p<(unsigned)NProcs; p++){
      for(int j=(ATOSEN[p]-1);j<(ATOSEN[p+1]-1);j++){
	
	unsigned haloNode = GATHER[j] - 1;
	unsigned unn = MFnode_list[haloNode].get_unn();
	
	// Store such that the halo is always in order of increasing unn.
	shared_unns[p].push_back( unn );
      }
    }
    
    //
    // Correct the UNN of the halo nodes (aka. SCATER nodes)
    //
    {    
      vector<unsigned> halo_cnt( NProcs );
      for(int i = 0; i<NProcs; i++)
	halo_cnt[i] = ATOREC[i+1] - ATOREC[i];
      
      vector< vector<unsigned> > halo_unns;
      halo_update(shared_unns, halo_cnt, halo_unns);
      
      // Write in the correct UNN's
      for(unsigned i=0; i<(unsigned)NProcs; i++){
	unsigned len = halo_unns[i].size();
	for(unsigned j = 0; j<len; j++){
	  unsigned haloNode = SCATER[ ATOREC[i]-1+j ];
	  MFnode_list[ haloNode ].set_unn( halo_unns[i][j] );
	}
      }
    }
    
    unsigned eid=0;    
    { // insert pressure node list into volume elements
      for(int e=0; e<NELM; e++){
	
	// Don't bother with elements that were deleted
	if(deleted_vol_elems.find(e) != deleted_vol_elems.end())
	  continue;
	
	int cnt = ENLBAS[e+1]-ENLBAS[e];
	int pos = ENLBAS[e];                                // Note: no -1 here
	vector<unn_t> nl(cnt);
	for(int n=0; n<cnt; n++){
	  int nodeNum = ENLIST[pos + n] - 1;
	  assert(nodeNum>=0   ); // or crap out
	  assert(nodeNum<NNOD ); // or crap out
	  nl[ n ] = MFnode_list[nodeNum].get_unn();
	}
	
	element_list[eid++].set_MFenlist( nl );
      }
    }
    
    //
    // Fluidity doesn't have surface pressure elements
    //
  }

  deleted_vol_elems.clear();
  deleted_sur_elems.clear();
}

void Mesh::import_vtu(vtkUnstructuredGrid *ug, std::vector<int> &offsets, int NPrivateNodes,
                      vector<int> &SENList, vector<int> &sids){
  deleted_vol_elems.clear();
  deleted_sur_elems.clear();

  dimension = 3;

  field_names.clear();

  int MyRank = MPI::COMM_WORLD.Get_rank();
  int NProcs = MPI::COMM_WORLD.Get_size();

  // Map nodal information to internal format.
  int flen = ug->GetPointData()->GetNumberOfArrays();
  int NNodes = ug->GetNumberOfPoints();
  assert(NPrivateNodes<=NNodes);

  // Get list of field names.
  for(int f=0;f<flen;f++){
    string fldname(ug->GetPointData()->GetArrayName(f));
    if(fldname=="GlobalNodeNumbers" || (fldname=="metric")){
      continue;
    }else{
      field_names.push_back(fldname);
    }
  }
  
  for(int n=0;n<NNodes;n++){
    SAM_Node node;
    
    node.set_gnn(n);
    node.set_current_owner(MyRank);
    
    // Nodewise field values.
    for(int f=0;f<flen;f++){
      string fldname(ug->GetPointData()->GetArrayName(f));
      if(fldname=="GlobalNodeNumbers" || (fldname=="metric")){
        continue;
      }else{
        node.append_field(ug->GetPointData()->GetArray(f)->GetTuple(n),
                          ug->GetPointData()->GetArray(f)->GetNumberOfComponents());
      }
    }
    
    // Position.
    double *X = ug->GetPoint(n);
    node.set_coord(X[0], X[1], X[2]);
    
    // Nodewise metric used for adaptivity. This is handled seperately
    // from fields as it can be used to control the
    // graph-partitioning.
    if(ug->GetPointData()->GetArray("metric")!=NULL){
      node.set_metric(ug->GetPointData()->GetArray("metric")->GetTuple(n), 9);
    }

    add_node(node);
  }
  
  unn_offsets = offsets;
  for(int i=0;i<NPrivateNodes;i++)
    node_list[i].set_unn(ug->GetPointData()->GetArray("GlobalNodeNumbers")->GetTuple1(i));

  vector< vector<int> > halo(NProcs);
  for(int i=NPrivateNodes;i<NNodes;i++){
    int global_nid = ug->GetPointData()->GetArray("GlobalNodeNumbers")->GetTuple1(i);
    node_list[i].set_unn(global_nid);

    int owner = -1;
    for(int p=0;p<NProcs;p++){
      if(global_nid<offsets[p+1]){
        owner = p;
        break;
      }
    }
    if(owner==-1){
      cerr<<"global_nid = "<<global_nid<<", "<<offsets[NProcs]<<endl;
      for(int p=0;p<=NProcs;p++){
        cerr<<"offsets["<<p<<"] = "<<offsets[p]<<endl;
      }
    }
    assert(owner!=-1);
      
    node_list[i].set_current_owner(owner);
    node_list[i].set_flags(NODE_HALO);
    
    halo[owner].push_back(global_nid);
  }
  do_node_headcount();
  find_nodes_per_rank( num_nodes("native") );
  assert(nodes_per_rank[MyRank]==NPrivateNodes);

  // Inform all process how much data they are receiving from every
  // other process. This is implemented using MPI_Alltoall because
  // some implementations of MPI are a bit restrictive on the number
  // of non-blocking communications that can be initiated.
  vector<int> send_count(NProcs), recv_count(NProcs);
  for(int p=0;p<NProcs;p++)
    send_count[p] = halo[p].size();
  MPI::COMM_WORLD.Alltoall(&(send_count[0]), 1, MPI::INT,
                           &(recv_count[0]), 1, MPI::INT);
  
  // MPI_Status sendStatus, recvStatus;
  vector<MPI::Request> sendRequest(NProcs, MPI::REQUEST_NULL);
  vector<MPI::Request> recvRequest(NProcs, MPI::REQUEST_NULL);
  MPI::Status recvStatus;
  
  // Send
  for(int p=0;p<NProcs;p++)
    if(send_count[p]){
      sendRequest[p] = MPI::COMM_WORLD.Isend(&(halo[p][0]), halo[p].size(), MPI::INT, p, 1);
    }
  // Receive
  vector< vector<int> > shared(NProcs);
  for(int p=0;p<NProcs;p++){
    if(recv_count[p]){
      shared[p].resize(recv_count[p]);
      recvRequest[p] = MPI::COMM_WORLD.Irecv(&(shared[p][0]), shared[p].size(), MPI::INT, p, 1);
    }
  }

  MPI::Request::Waitall(NProcs, &(sendRequest[0])); // Wait all sends
  MPI::Request::Waitall(NProcs, &(recvRequest[0])); // Wait all receives
  
  halo_nodes.resize(NProcs);
  shared_nodes.resize(NProcs);
  for(int p=0;p<NProcs;p++){
    for(vector<int>::const_iterator it=halo[p].begin();it!=halo[p].end();++it){
      halo_nodes[p].insert(*it);
    }
    for(vector<int>::const_iterator it=shared[p].begin();it!=shared[p].end();++it){
      shared_nodes[p].insert(*it);
    }
  }

  // Convert volume element information
  unsigned eid=0;
  int NElements = ug->GetNumberOfCells();
  SAM_Element elem;
  for(int e=0;e<NElements;e++){
    elem.set_flags(ELM_DEFAULT|ELM_VOLUME);
    bool wanted = false;
    
    { // write element node list with UNN's
      vector<unn_t> nl(4);
      for(int n=0;n<4;n++){
        int nid = ug->GetCell(e)->GetPointId(n);
        nl[n] = node_list[nid].get_unn();
        
        if(node_list[nid].get_current_owner()==MyRank)
          wanted = true;
      }
      elem.set_enlist( nl );
    }
      
    if(!wanted){
      deleted_vol_elems.insert(e);
      cout<<"deleted elements "<<e<<endl;
      continue;
    }
    
    elem.set_eid(eid++); 
    add_element( elem );
  }

  int snloc=3;
  int NSElements=sids.size();
  { // Import surface.
    ECHO( "Import surface." );
    
    SAM_Element elem;
    for(int e=0; e<NSElements; e++){
      elem.set_flags(ELM_DEFAULT|ELM_SURFACE);
      bool wanted = false; // Maybe it's on the halo interface

      { // write element node list with UNN's
	int cnt = snloc;
	int pos = e*snloc;
	vector<unn_t> nl(cnt);
	for(int n=0; n<cnt; n++){
	  int nodeNum = SENList[pos + n];
	  assert(nodeNum>=0);
	  assert(nodeNum<NNodes);
	  nl[n] = node_list[nodeNum].get_unn();
          if(node_list[nodeNum].get_current_owner() == MyRank)
            wanted = true;
	}
        if(!wanted){
	  deleted_sur_elems.insert(e);
          continue;
	}
        
	elem.set_enlist(nl);
      }
      
      elem.set_eid(eid++);

      // add surface id's
      vector<int> ifield(1);
      ifield[0] = sids[e];
      elem.set_ifields(ifield);

      // Field values??
      ECHO( "Adding element:");
      CHECK( elem );
      
      add_element(elem);
    }
  }

  do_element_headcount();
  
  return;
}
