/* Copyright (C) 2006 Imperial College London and others.
   
   Please see the AUTHORS file in the main source directory for a full list
   of copyright holders.
   
   Applied Modelling and Computation Group
   Department of Earth Science and Engineering
   Imperial College London
   
   AMCGSoftware@imperial.ac.uk
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
   USA
*/

//
// Explination of sam options
//

// options[0]:  m // Target number of partitions. If this is 0, then
//                // the number of ranks in MPI_COMM_WORLD is assumed.
// options[1]:  1 // clean repartitioning into nparts
//              2 // re-partitioning diffusive
//              3 // re-partitioning (repartition-remapping)
// options[2]:  decrepit

// options[3]:  1 // no node weights
//              2 // calculate edge weights based on the projected 
//                   density of nodes in the future
// options[4]:  1 // no edge weights
//              2 // calculate edge weights optimizing for adaptivity
//              3 // calculate edge weights based on curvature of fields
// options[5]:  1 // Level 1 halo
//              2 // Level 2 halo
#ifdef HAVE_VTK
#include "vtu_sam.h"

using namespace std;
 
void sam_init(vtkUnstructuredGrid *ug, int NPrivateNodes, vector<int> &SENList, vector<int> &sids){
#ifdef HAVE_MPI
  int options[] = {0, 4, 0, 1, 1, 1};
  samfloat_t fxnl_tol = 0;

  if(mesh != NULL){
    delete mesh;
    mesh = NULL;
  }
  mesh = new Mesh;
  
  mesh->mixed_formulation(false);
  
  int nparts = MPI::COMM_WORLD.Get_size();
  vector<int> offsets(nparts+1, 0);
  MPI::COMM_WORLD.Allgather(&NPrivateNodes, 1, MPI::INT, &(offsets[1]), 1, MPI::INT);
  for(int i=1;i<=nparts;i++){
    offsets[i] += offsets[i-1];
  }

  mesh->import_vtu(ug, offsets, NPrivateNodes, SENList, sids);
  
  // Build options.
  decomp_opts.resize(5);
  for(int i=0;i<5;i++)
    decomp_opts[i] = options[i];
  
  mesh->set_functional_tol(fxnl_tol);
#endif
  return;
}
#endif

void sam_migrate(bool diffusion){
#ifdef HAVE_MPI
  assert(mesh != NULL);
  if(diffusion){
    decomp_opts[1] = 2;
    decomp_opts[4] = 2;
  }else{
    decomp_opts[1] = 3;
    decomp_opts[4] = 1;
  }

  // Graph partitioning.
  vector<int> noddom = mesh->decomp(decomp_opts);
  
  // Migrate Mesh according to noddom.
  mesh->migrate(noddom);
#endif
  return;
}

#ifdef HAVE_VTK
vtkUnstructuredGrid *sam_export_mesh(int &NPrivateNodes, vector<int> &SENList, vector<int> &sids){
#ifdef HAVE_MPI
  assert(mesh != NULL);

  vtkUnstructuredGrid *ug = vtkUnstructuredGrid::New();

  int NProcs = MPI::COMM_WORLD.Get_size();
  int MyRank = MPI::COMM_WORLD.Get_rank();
  
  NPrivateNodes = mesh->node_list.psize();
    
  int NNodes = mesh->node_list.size();
  map<int, int> unn2unn;
  mesh->renumber(unn2unn);

  vtkPoints *points = vtkPoints::New();
  vtkIntArray *GlobalNodeNumbers = vtkIntArray::New();
  
  GlobalNodeNumbers->SetNumberOfTuples(NNodes);
  GlobalNodeNumbers->SetNumberOfComponents(1);
  GlobalNodeNumbers->SetName("GlobalNodeNumbers");

  map<string, vtkDoubleArray *> fields;
  for(vector<string>::iterator name=mesh->field_names.begin();name!=mesh->field_names.end();++name){
    fields[*name] = vtkDoubleArray::New();
    fields[*name]->SetNumberOfTuples(NNodes);
    fields[*name]->SetNumberOfComponents(1);
    fields[*name]->SetName((*name).c_str());
  }
  vtkDoubleArray *metric = vtkDoubleArray::New();
  metric->SetNumberOfComponents(9);
  metric->SetNumberOfTuples(NNodes);
  metric->SetName("metric");

  points->SetNumberOfPoints(NNodes);
  int nfields=0;
  if(NNodes>0)
    nfields = mesh->node_list[0].get_size_fields();
  for(int i=0; i<NNodes; i++){
    points->SetPoint(i, mesh->node_list[i].get_x(), mesh->node_list[i].get_y(), mesh->node_list[i].get_z());
    for(int j=0;j<nfields;++j){
      fields[mesh->field_names[j]]->SetTuple1(i, mesh->node_list[i].get_fields()[j]);
    }
    GlobalNodeNumbers->SetTuple1(i, mesh->get_node_unn(i));
    assert(mesh->node_list[i].get_size_metric()==9);
    metric->SetTuple(i, mesh->node_list[i].get_cptr_metric());
  }
  ug->SetPoints(points);
  ug->Update();
  points->Delete();

  for(map<string, vtkDoubleArray *>::iterator it=fields.begin();it!=fields.end();++it){
    ug->GetPointData()->AddArray(it->second);
    ug->Update();
    it->second->Delete();
  }
  if(ug->GetPointData()->GetArray("GlobalNodeNumbers")!=NULL)
    ug->GetPointData()->RemoveArray("GlobalNodeNumbers");
  ug->Update();
  
  ug->GetPointData()->AddArray(GlobalNodeNumbers);
  ug->Update();
  GlobalNodeNumbers->Delete();

  if(ug->GetPointData()->GetArray("metric")!=NULL)
    ug->GetPointData()->RemoveArray("metric");
  ug->GetPointData()->AddArray(metric);
  ug->Update();
  metric->Delete();

  if(ug->GetPointData()->GetArray("NodeOwner")!=NULL)
    ug->GetPointData()->RemoveArray("NodeOwner");
  
  vector<int> nowner(NNodes);
  if(NProcs>1){
    vtkIntArray *nowner_vtk = vtkIntArray::New();
    nowner_vtk->SetNumberOfComponents(1);
    nowner_vtk->SetNumberOfTuples(NNodes);
    nowner_vtk->SetName("NodeOwner");
    for(int i=0;i<NNodes;i++)
      nowner_vtk->SetTuple1(i, MyRank);
    
    for(int i=0;i<NNodes;i++){
      nowner[i] = mesh->get_node_future_owner(i);
      nowner_vtk->SetTuple1(i, mesh->get_node_future_owner(i));
    }
    ug->GetPointData()->AddArray(nowner_vtk);
    ug->Update();
    nowner_vtk->Delete();
  }

  int num_elements = mesh->element_list.size();
  SENList.clear();
  sids.clear();
  vector<int> vec_eowner;
  for(int i=0;i<num_elements;i++){
    unsigned char type = mesh->element_list[i].get_flags();
    if( type & ELM_SURFACE ){
      vector<unn_t> enl( mesh->element_list[i].get_enlist() );
      const vector<int>& ifields = mesh->element_list[i].get_ifields();
      for(int j=0;j<3;j++)
        SENList.push_back(mesh->unn2gnn(unn2unn[enl[j]]));
      sids.push_back(ifields[0]);
    }else{
      vector<unn_t> lcell(mesh->element_list[i].get_enlist());
      vtkIdType cell[4];
      int owner=MyRank;
      for(int j=0;j<4;j++){
        int gnn = mesh->unn2gnn(unn2unn[lcell[j]]);
        owner = min(owner, nowner[gnn]);
        cell[j] = gnn;
      }
      vec_eowner.push_back(owner);
      ug->InsertNextCell(VTK_TETRA, 4, cell);
    }
  }
  
  num_elements = ug->GetNumberOfCells();

  if(NProcs>1){
    vtkIntArray *eowner = vtkIntArray::New();
    eowner->SetNumberOfComponents(1);
    eowner->SetNumberOfTuples(num_elements);
    eowner->SetName("ElementOwner");
    for(int e=0;e<num_elements;e++){
      eowner->SetTuple1(e, vec_eowner[e]);
    }
    ug->GetCellData()->AddArray(eowner);
    ug->Update();
    eowner->Delete();
  }

  return ug;
#else
  return NULL;
#endif
}
#endif
  
void sam_cleanup(){
#ifdef HAVE_MPI
  if(mesh != NULL){
    delete mesh;
    mesh = NULL;
    
    decomp_opts.clear();
  }
#endif
  return;
}
