/*
  Copyright (C) 2006 Imperial College London and others.

  Please see the AUTHORS file in the main source directory for a full list
  of copyright holders.

  Applied Modelling and Computation Group
  Department of Earth Science and Engineering
  Imperial College London

  AMCGSoftware@imperial.ac.uk

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
  USA
*/

#include "DiscreteGeometryConstraints.h"

using namespace std;

bool DiscreteGeometryConstraints::verbose=false;


DiscreteGeometryConstraints::DiscreteGeometryConstraints(){
  if(verbose)
    cout<<"DiscreteGeometryConstraints::DiscreteGeometryConstraints()\n";
  set_coplanar_tolerance(0.9999999);
  ug = NULL;
}

void DiscreteGeometryConstraints::set_coplanar_tolerance(double tol){
  if(verbose)
    cout<<"void DiscreteGeometryConstraints::set_coplanar_tolerance("<<tol<<")\n";
  COPLANAR_MAGIC_NUMBER = tol;
}

void DiscreteGeometryConstraints::verbose_off(){
  verbose = false;
}

void DiscreteGeometryConstraints::verbose_on(){
  verbose = true;
}

void DiscreteGeometryConstraints::get_constraints(vector<double> &max_len){
  if(verbose)
    cout<<"void DiscreteGeometryConstraints::get_constraints(vector<double> &max_len)\n";

  if(gconstraint.empty())
    find_geometry_constraints();

  size_t NNodes = ug->GetNumberOfPoints();
  max_len.resize(9*NNodes);
  for(size_t i=0; i<9*NNodes; i++)
    max_len[i] = gconstraint[i];

  return;
}

void DiscreteGeometryConstraints::get_surface(std::vector<int> &ENList){
  if(verbose)
    cout<<"void DiscreteGeometryConstraints::get_surface(vector<int> &ENList)\n";
  ENList = SENList;

  return;
}

void DiscreteGeometryConstraints::set_volume_input(vtkUnstructuredGrid *_ug, int NPrivateNodes){
  if(verbose)
    cout<<"void DiscreteGeometryConstraints::set_volume_input(vtkUnstructuredGrid *ug)\n";
  ug = _ug;
  ug->Update();

#ifdef HAVE_MPI
  int NProcs = MPI::COMM_WORLD.Get_size();
  int MyRank = MPI::COMM_WORLD.Get_rank();
#else
  int NProcs = 1;
  int MyRank = 0;
#endif

  // Get node-element adjancy list
  size_t NElements = ug->GetNumberOfCells();
  size_t NNodes = ug->GetNumberOfPoints();
  std::deque< std::set<size_t> > NEList(NNodes);
  for(size_t i=0;i<NElements;i++){
    vtkTetra *tetra = (vtkTetra *)ug->GetCell(i);
    for(size_t j=0;j<4;j++){
      NEList[tetra->GetPointId(j)].insert(i);
    }
  }

  // Search for outter surface
  for(size_t i=0;i<NElements;i++){
    vtkTetra *tetra = (vtkTetra *)ug->GetCell(i);
    for(size_t j=0;j<4;j++){
      bool found=false;
      size_t nid1, nid2, nid3;
      if(j==0){
        nid1=tetra->GetPointId(1);
        nid2=tetra->GetPointId(3);
        nid3=tetra->GetPointId(2);
      }else if(j==1){
        nid1=tetra->GetPointId(0);
        nid2=tetra->GetPointId(2);
        nid3=tetra->GetPointId(3);
      }else if(j==2){
        nid1=tetra->GetPointId(0);
        nid2=tetra->GetPointId(3);
        nid3=tetra->GetPointId(1);
      }else{
        nid1=tetra->GetPointId(0);
        nid2=tetra->GetPointId(1);
        nid3=tetra->GetPointId(2);
      }
      
      for(set<size_t>::iterator it=NEList[nid1].begin();it!=NEList[nid1].end();++it){
        if((*it)==i){
          continue;
        }
        
        if(NEList[nid2].find(*it)!=NEList[nid2].end()){
          if(NEList[nid3].find(*it)!=NEList[nid3].end()){
            found=true;
            break;
          }
        }
      }
      if(!found && (nid1<NPrivateNodes || nid2<NPrivateNodes || nid3<NPrivateNodes)){
        SENList.push_back(nid1);
        SENList.push_back(nid2);
        SENList.push_back(nid3);
      }
    }
  }

  int NFacets = SENList.size()/3;
  get_coplanar_ids();

  if(NProcs>1){
    set<int> fids;
    for(vector<int>::iterator it=coplanar_ids.begin();it!=coplanar_ids.end();++it)
      fids.insert(*it);
    int nfids = fids.size();
    
    vector<int> foffsets(NProcs+1);
    MPI::COMM_WORLD.Allgather(&nfids, 1, MPI::INT, &(foffsets[1]), 1, MPI::INT);
    foffsets[0] = 0;
    for(int i=1;i<=NProcs;i++){
      foffsets[i] += foffsets[i-1];
    }
    for(vector<int>::iterator it=coplanar_ids.begin();it!=coplanar_ids.end();++it)
      *it += foffsets[MyRank];
    
    vector<int> offsets(NProcs+1, 0);
    MPI::COMM_WORLD.Allgather(&NPrivateNodes, 1, MPI::INT, &(offsets[1]), 1, MPI::INT);
    for(int i=1;i<=NProcs;i++){
      offsets[i] += offsets[i-1];
    }
    map<set<int>, int> halo_surface_recv;
    vector< vector<int> > halo_surface_send(NProcs);
    vector<int> recv_cnt(NProcs, 0);
    for(int i=0;i<NFacets;++i){
      set<int> neigh;
      for(int j=0;j<3;j++){
        int lnn = SENList[i*3+j];
        if(lnn>=NPrivateNodes){
          int gnn = ug->GetPointData()->GetArray("GlobalNodeNumbers")->GetTuple1(lnn);
          for(int p=0;p<NProcs;p++){
            if(gnn<offsets[p+1]){
              neigh.insert(p);
              break;
            }
          }
        }
      }
      
      if(neigh.size()==0)
        continue;
      
      if(*neigh.begin()<MyRank){
        set<int> gfacet;
        for(int j=0;j<3;j++)
          gfacet.insert(ug->GetPointData()->GetArray("GlobalNodeNumbers")->GetTuple1(SENList[i*3+j]));;
        halo_surface_recv[gfacet] = coplanar_ids[i];
        recv_cnt[*neigh.begin()]++;
      }else{ // MyRank must be owner.
        for(set<int>::iterator it=neigh.begin();it!=neigh.end();++it){
          int n0 = ug->GetPointData()->GetArray("GlobalNodeNumbers")->GetTuple1(SENList[i*3  ]);
          int n1 = ug->GetPointData()->GetArray("GlobalNodeNumbers")->GetTuple1(SENList[i*3+1]);
          int n2 = ug->GetPointData()->GetArray("GlobalNodeNumbers")->GetTuple1(SENList[i*3+2]);
          
          halo_surface_send[*it].push_back(n0);
          halo_surface_send[*it].push_back(n1);
          halo_surface_send[*it].push_back(n2);
          halo_surface_send[*it].push_back(coplanar_ids[i]);
        }
      }
    }
    
    // MPI_Status sendStatus, recvStatus;
    vector<MPI::Request> sendRequest(NProcs, MPI::REQUEST_NULL);
    vector<MPI::Request> recvRequest(NProcs, MPI::REQUEST_NULL);
    MPI::Status recvStatus;
    
    // Send
    for(int p=0;p<NProcs;p++){
      if(halo_surface_send[p].size()){
        sendRequest[p] = MPI::COMM_WORLD.Isend(&(halo_surface_send[p][0]), halo_surface_send[p].size(), MPI::INT, p, 2);
      }
    }
    // Receive
    vector< vector<int> > recv_buffer(NProcs);
    for(int p=0;p<NProcs;p++){
      if(recv_cnt[p]){
        recv_buffer[p].resize(recv_cnt[p]*4);
        recvRequest[p] = MPI::COMM_WORLD.Irecv(&(recv_buffer[p][0]), recv_buffer[p].size(), MPI::INT, p, 2);
      }
    }
      
    MPI::Request::Waitall(NProcs, &(sendRequest[0])); // Wait all sends
    MPI::Request::Waitall(NProcs, &(recvRequest[0])); // Wait all receives
    
    map<int, int> colour_map;
    for(int p=0;p<NProcs;p++){
      for(int i=0;i<recv_cnt[p];i++){
        set<int> gfacet;
        gfacet.insert(recv_buffer[p][i*4]);
        gfacet.insert(recv_buffer[p][i*4+1]);
        gfacet.insert(recv_buffer[p][i*4+2]);

        int owner_patch_id = recv_buffer[p][i*4+3];
        int local_patch_id = halo_surface_recv[gfacet];
        if(colour_map.count(local_patch_id)){
          colour_map[local_patch_id] = min(colour_map[local_patch_id], owner_patch_id);
        }else{
          colour_map[local_patch_id] = owner_patch_id;
        }
      }
    }
    
    for(vector<int>::iterator it=coplanar_ids.begin();it!=coplanar_ids.end();++it){
      if(colour_map.count(*it))
        *it = colour_map[*it];
    }
  }

  return;
}

void DiscreteGeometryConstraints::set_surface_input(vtkUnstructuredGrid *_ug, vector<int> &_SENList, vector<int> &ids){
  if(verbose)
    cout<<"void DiscreteGeometryConstraints::set_surface_input(vtkUnstructuredGrid *ug, vector<int> &_SENList, vector<int> &ids)\n";
  ug = _ug;
  SENList = _SENList;
  
  // Create EEList for surface
  size_t NSElements = SENList.size()/3;
  for(size_t i=0;i<NSElements;i++){
    for(size_t j=0;j<3;j++){
      SNEList[SENList[3*i+j]].insert(i);
    }
  }
  
  coplanar_ids = ids;

  return;
}

#ifdef HAVE_VTK
void DiscreteGeometryConstraints::write_vtk(std::string filename){
  if(verbose)
    cout<<"void DiscreteGeometryConstraints::write_vtk(std::string filename)\n";
  
  vtkUnstructuredGrid *sug = vtkUnstructuredGrid::New();
  sug->SetPoints(ug->GetPoints());

  size_t NSElements = SENList.size()/3;
  for(size_t i=0;i<NSElements;i++){
    vtkIdType pts[3];
    for(size_t j=0;j<3;j++){
      pts[j] = SENList[i*3+j];
    }
    sug->InsertNextCell(VTK_TRIANGLE, 3, pts);
  }

  if(coplanar_ids.size()){
    assert(coplanar_ids.size()==NSElements);
    vtkIntArray *field = vtkIntArray::New();
    field->SetNumberOfComponents(1);
    field->SetNumberOfTuples(NSElements);
    field->SetName("coplanar_ids");
    for(size_t i=0;i<NSElements;i++){
      field->SetTuple1(i, coplanar_ids[i]);
    }
    sug->GetCellData()->AddArray(field);
    sug->Update();
    field->Delete();
  }

  size_t NNodes = sug->GetNumberOfPoints();
  if(gconstraint.size()){
    vtkDoubleArray *field = vtkDoubleArray::New();
    field->SetNumberOfComponents(9);
    field->SetNumberOfTuples(NNodes);
    field->SetName("max_lengths");
    for(size_t i=0;i<NNodes;i++)
      field->SetTuple(i, &(gconstraint[i*9]));
    sug->GetPointData()->AddArray(field);
    sug->Update();
    field->Delete();
  }

  vtkXMLUnstructuredGridWriter *sug_writer = vtkXMLUnstructuredGridWriter::New();
  sug_writer->SetFileName(filename.c_str());
  sug_writer->SetInput(sug);
  sug_writer->Write();
  sug_writer->Delete();
  
  sug->Delete();
  return;
}
#endif

void DiscreteGeometryConstraints::normal(int v1, int v2, int v3, double *n){
  double r1[3], r2[3], r3[3];
  ug->GetPoint(v1, r1);
  ug->GetPoint(v2, r2);
  ug->GetPoint(v3, r3);
    
  double ax = r3[0] - r2[0];
  double ay = r3[1] - r2[1];
  double az = r3[2] - r2[2];

  double bx = r1[0] - r2[0];
  double by = r1[1] - r2[1];
  double bz = r1[2] - r2[2];
  
  n[0] = (ay * bz - az * by);
  n[1] = (az * bx - ax * bz);
  n[2] = (ax * by - ay * bx);

  double length = sqrt(n[0]*n[0]+n[1]*n[1]+n[2]*n[2]);
  
  n[0]/=length;
  n[1]/=length;
  n[2]/=length;
  
  return;
}

void DiscreteGeometryConstraints::get_coplanar_ids(){
  if(verbose)
    cout<<"void DiscreteGeometryConstraints::get_coplanar_ids()\n";

  // Calculate all element normals
  size_t NSElements = SENList.size()/3;
  vector<double> normals(NSElements*3);
  for(size_t i=0;i<NSElements;i++){
    normal(SENList[3*i], SENList[3*i+1], SENList[3*i+2], &(normals[i*3]));
  }

  // Create EEList for surface
  for(size_t i=0;i<NSElements;i++){
    for(size_t j=0;j<3;j++){
      SNEList[SENList[3*i+j]].insert(i);
    }
  }
  
  EEList.resize(NSElements*3);
  for(size_t i=0;i<NSElements;i++){
    for(size_t j=0;j<3;j++){
      size_t nid1=SENList[i*3+(j+1)%3];
      size_t nid2=SENList[i*3+(j+2)%3];
      for(set<size_t>::iterator it=SNEList[nid1].begin();it!=SNEList[nid1].end();++it){
        if((*it)==i){
          continue;
        }
        
        if(SNEList[nid2].find(*it)!=SNEList[nid2].end()){
          EEList[i*3+j] = *it;
          break;
        }
      }
    }
  }

  // Form patches
  coplanar_ids.resize(NSElements);
  for(vector<int>::iterator it=coplanar_ids.begin(); it!=coplanar_ids.end(); ++it)
    *it = 0;
  
  size_t current_id = 1;
  for(size_t pos = 0;pos<NSElements;){
    // Create a new starting point
    double *ref_normal=NULL;
    for(size_t i=pos;i<NSElements;i++){{
        if(coplanar_ids[i]==0){
          // This is the first element in the new patch
          pos = i;
          coplanar_ids[pos] = current_id;
          ref_normal = &(normals[pos*3]);
          break;
        }
      }
      
      // Jump out of this while loop if we are finished
      if(i==NSElements)
        break;
    }
    
    // Initialise the front
    set<int> front;
    front.insert(pos);
          
    // Advance this front
    while(!front.empty()){
      int sele = *front.begin();
      front.erase(front.begin());

      // Check surrounding surface elements:      
      for(int i=0; i<3; i++){
        int sele2 = EEList[sele*3+i];
        
        if(coplanar_ids[sele2]>0)
          continue;
          
        double coplanar =
          ref_normal[0]*normals[sele2*3  ]+
          ref_normal[1]*normals[sele2*3+1]+
          ref_normal[2]*normals[sele2*3+2];

        if(coplanar>=COPLANAR_MAGIC_NUMBER){
          front.insert(sele2);
          coplanar_ids[sele2] = current_id;
        }
      }
    }
    current_id++;
    pos++;
  }

  return;
}

void DiscreteGeometryConstraints::get_coplanar_ids(std::vector<int> &ids){
  if(verbose)
    cout<<"void get_coplanar_ids(std::vector<int> &ids)\n";

  if(coplanar_ids.empty())
    get_coplanar_ids();

  ids = coplanar_ids;
}

void DiscreteGeometryConstraints::find_geometry_constraints(){
  if(verbose)
    cout<<"void DiscreteGeometryConstraints::find_geometry_constraints()\n";

  // Find the maximum lengths that will be used for constraining
  // the metric.
  double bbox[6];
  ug->GetBounds(bbox);
  double max_dist = max(max(bbox[1]-bbox[0], bbox[3]-bbox[2]), bbox[5]-bbox[4]);

  if(verbose)
    cout<<"Bounding box = ("
        <<bbox[0]<<", "<<bbox[1]<<"), ("
        <<bbox[2]<<", "<<bbox[3]<<"), ("
        <<bbox[4]<<", "<<bbox[5]<<")"<<endl;

  // Initialise geometry constraints
  size_t NNodes = ug->GetNumberOfPoints();
  gconstraint.resize(9*NNodes);

  for(size_t i=0;i<NNodes;i++){
    gconstraint[i*9  ] = max_dist; gconstraint[i*9+1] = 0.0;      gconstraint[i*9+2] = 0.0; 
    gconstraint[i*9+3] = 0.0;      gconstraint[i*9+4] = max_dist; gconstraint[i*9+5] = 0.0; 
    gconstraint[i*9+6] = 0.0;      gconstraint[i*9+7] = 0.0;      gconstraint[i*9+8] = max_dist;
  }

  size_t NSElements=coplanar_ids.size();
  set<int> patch_ids;
  for(size_t i=0;i<NSElements;i++)
    patch_ids.insert(coplanar_ids[i]);

  // Loop over each patch, merging in the constraints from each patch
  // into the nodes that lie along geometry edges.
  for(set<int>::iterator ip=patch_ids.begin();ip!=patch_ids.end();++ip){
    int p = *ip;
    // Create an node-element adjancy list, and the list of elements
    // for this patch.
    map<size_t, set<size_t> > patch_nelist;
    size_t patch_nelements;
    vector<size_t> patch_elements;
    set<int> patch_nodes;
    {
      for(size_t ele=0;ele<NSElements;ele++){
        if(coplanar_ids[ele]==(int)p){
          patch_elements.push_back(ele);
          for(size_t i=0;i<3;i++){
            patch_nelist[SENList[ele*3+i]].insert(ele);
            patch_nodes.insert(SENList[ele*3+i]);
          }
        }
      }

      patch_nelements = patch_elements.size();
    }

    assert(patch_nelements>=0);

    // Create element-element adjancy list for this patch.
    vector<int> patch_eelist(patch_nelements*3, -1);
    vector<size_t> patch_boundary_edges;
    set<size_t> patch_boundary_nodes;
    for(size_t i=0;i<patch_nelements;i++){
      for(size_t j=0;j<3;j++){
        size_t n1 = SENList[patch_elements[i]*3+(j+1)%3];
        size_t n2 = SENList[patch_elements[i]*3+(j+2)%3];

        for(set<size_t>::const_iterator it=patch_nelist[n1].begin(); it!=patch_nelist[n1].end();++it){
          set<size_t>::const_iterator jt=patch_nelist[n2].find(*it);
          if(jt!=patch_nelist[n2].end() && (*jt!=patch_elements[i])){
            patch_eelist[i*3 + j] = *it;
            break;
          }
        }

        if(patch_eelist[i*3 + j]==-1){
          patch_boundary_nodes.insert(n1);
          patch_boundary_nodes.insert(n2);
          patch_boundary_edges.push_back(n1);
          patch_boundary_edges.push_back(n2);
        }
      }
    }

    // Identify the corner nodes of this patch.
    map<size_t, set<size_t> > corner;
    for(set<size_t>::const_iterator it=patch_boundary_nodes.begin();it!=patch_boundary_nodes.end();++it){
      set<size_t> patches;
      for(set<size_t>::iterator jt=SNEList[*it].begin();jt!=SNEList[*it].end();++jt){
        patches.insert(coplanar_ids[*jt]);
      }
      
      if(patches.size()>2){
        corner[*it] = patches;
      }
    }
    
    // Sort edges to form polygon.
    assert(patch_boundary_edges.size()!=0);
    {
      // First ensure that a patch boundary starts at a corner.
      vector<size_t>::iterator it=patch_boundary_edges.begin();
      for(;it!=patch_boundary_edges.end();){
        vector<size_t>::iterator first = it++;
        vector<size_t>::iterator second = it++;
        if(corner.count(*first)){
          swap_ranges(first, ++second, patch_boundary_edges.begin());
          break;
        }
      }
      
      it=patch_boundary_edges.begin(); ++it;
      size_t end_node = *it; it++;
      vector<size_t>::iterator tail = it;
      while(it!=patch_boundary_edges.end()){
        vector<size_t>::iterator first = it++;
        vector<size_t>::iterator second = it++;
        // Note that the edges are consistently ordered because the
        // elements were consistently ordered.
        if(*first == end_node){
          end_node = *second;
          swap_ranges(first, (second+1), tail);
          tail += 2;
          it = tail;
        }
      }
    }

    // Split polygon into sets of boundary lines.
    vector< pair<int, int> > patch_boundary_lines(corner.size());
    {
      size_t line=0;
      patch_boundary_lines[line].first = patch_boundary_edges[0];
      for(vector<size_t>::iterator it=patch_boundary_edges.begin();it!=patch_boundary_edges.end();it+=2){
        if(corner.count(*(it+1))){
          patch_boundary_lines[line].second = *(it+1);
          line++;
          if(line<=corner.size())
            patch_boundary_lines[line].first = *(it+1);
        }
      }
    }
    
    // Get the normal to the patch. This is also one of the
    // eigenvectors of the constraint tensor.
    double v1[3];
    {
      double r0[3], r1[3], r2[3];
      ug->GetPoint(SENList[patch_elements[0]*3  ], r0);
      ug->GetPoint(SENList[patch_elements[0]*3+1], r1);
      ug->GetPoint(SENList[patch_elements[0]*3+2], r2);
      
      double a=r1[0]-r0[0];
      double b=r1[1]-r0[1];
      double c=r1[2]-r0[2];
      
      double d=r2[0]-r0[0];
      double e=r2[1]-r0[1];
      double f=r2[2]-r0[2];
      
      v1[0] = (b*f-c*e);
      v1[1] = -(a*f-d*c);
      v1[2] = (a*e-d*b);
      
      double norm = sqrt(v1[0]*v1[0]+v1[1]*v1[1]+v1[2]*v1[2]);
      
      v1[0] /= norm;
      v1[1] /= norm;
      v1[2] /= norm;
    }

    double v0[3];
    {
      double r0[3];
      ug->GetPoint(patch_boundary_lines[0].first, r0);
      
      double r1[3];
      ug->GetPoint(patch_boundary_lines[0].second, r1);
      
      v0[0] = r0[0] - r1[0];
      v0[1] = r0[1] - r1[1];
      v0[2] = r0[2] - r1[2];
      
      double norm = sqrt(v0[0]*v0[0]+v0[1]*v0[1]+v0[2]*v0[2]);
      
      v0[0] /= norm;
      v0[1] /= norm;
      v0[2] /= norm;
    }

    double min_dist = max_dist;
    for(vector< pair<int, int> >::iterator il=patch_boundary_lines.begin();il!=patch_boundary_lines.end();++il){
      assert((il->first)!=(il->second));
      double r0[3];
      ug->GetPoint(il->first, r0);
      
      double r1[3];
      ug->GetPoint(il->second, r1);
      
      double dx = r0[0] - r1[0];
      double dy = r0[1] - r1[1];
      double dz = r0[2] - r1[2];
      
      min_dist = min(min_dist, sqrt(dx*dx+dy*dy+dz*dz));
    }
    
    // Get the 3rd perpendicular eigenvector.
    double v2[3];
    {
      double a=v1[0];
      double b=v1[1];
      double c=v1[2];
      
      double d=v0[0];
      double e=v0[1];
      double f=v0[2];
      
      v2[0] = (b*f-c*e);
      v2[1] = -(a*f-d*c);
      v2[2] = (a*e-d*b);
      
      double norm = sqrt(v2[0]*v2[0]+v2[1]*v2[1]+v2[2]*v2[2]);
      
      v2[0] /= norm;
      v2[1] /= norm;
      v2[2] /= norm;
    }
      
    // Form metric tensor.
    double D[3], V[9];
    D[0] = min_dist;
    D[1] = min_dist;
    D[2] = max_dist;
    
    V[0] = v0[0]; V[1] = v0[1]; V[2] = v0[2];
    V[3] = v1[0]; V[4] = v1[1]; V[5] = v1[2];
    V[6] = v2[0]; V[7] = v2[1]; V[8] = v2[2];

    for(set<int>::iterator in=patch_nodes.begin();in!=patch_nodes.end();++in){
      for(int i=0;i<3;i++){
        for(int j=0;j<3;j++){
          gconstraint[(*in)*9+i*3+j] = 0.0;
          for(int k=0;k<3;k++){
            gconstraint[(*in)*9+i*3+j]+=D[k]*V[k*3+i]*V[k*3+j];
          }
        }
      }
    }
  }

  return;
}
