/* Copyright (C) 2009 Imperial College London.

 Please see the AUTHORS file in the main source directory for a full
 list of copyright holders.

 Dr Gerard J Gorman
 Applied Modelling and Computation Group
 Department of Earth Science and Engineering
 Imperial College London

 g.gorman@imperial.ac.uk

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 USA
*/
#include <algorithm>
#include <cmath>
#include <iostream>
#include <vector>

#include "confdefs.h"
#include "vtk.h"

#include "ErrorMeasure.h"
#include "Adaptivity.h"
#include "DiscreteGeometryConstraints.h"
#include "vtu_sam.h"

/** This test seeks to execute a full adaptive mesh example.
 */

using namespace std;

int main(int argc, char **argv){
  MPI::Init(argc,argv);
  
  int rank = MPI::COMM_WORLD.Get_rank();
  int NProcs = MPI::COMM_WORLD.Get_size();

  char filename[1024];

  sprintf(filename, "coarse_slab_with_data/coarse_slab_with_data_%d.vtu", rank);
  int NPrivateNodes[] = {333, 333, 333, 332};

  vtkXMLUnstructuredGridReader *ug_reader = vtkXMLUnstructuredGridReader::New();
  ug_reader->SetFileName(filename);
  ug_reader->Update();
  
  vtkUnstructuredGrid *ug = ug_reader->GetOutput();
  ug->Update();
  
  // These should be populated only once at the outset of a simulation
  // and be maintained thereafter.
  vector<int> SENList, sids;
  {
    DiscreteGeometryConstraints constraints;
    // constraints.verbose_on();
    constraints.set_coplanar_tolerance(0.9999999);
    constraints.set_volume_input(ug, NPrivateNodes[rank]);
    
    constraints.get_coplanar_ids(sids);
    constraints.get_surface(SENList);

    assert(sids.size()*3==SENList.size());

    sprintf(filename, "surface_%d.vtu", rank);
    constraints.write_vtk(filename);
  }

  DiscreteGeometryConstraints constraints;
  // constraints.verbose_on();
  constraints.set_surface_input(ug, SENList, sids);
  
  // vector<double> max_len;
  // constraints.get_constraints(max_len);
  // constraints.write_vtk(string("sids.vtu"));
  
  /* Test merging of metrics.
   */
  ErrorMeasure error;
  // error.verbose_on();
  error.set_input(ug, NPrivateNodes[rank]);

  error.add_field("Vm", 0.1, true, 0.01);
  error.set_min_length(0.002);
  error.apply_gradation(1.3);
 
  error.set_max_nodes(100000);  
  error.set_max_length(0.04);
  
  error.diagnostics();
  
  vtkXMLUnstructuredGridWriter *metric_writer = vtkXMLUnstructuredGridWriter::New();
  sprintf(filename, "metric_%d.vtu", rank);
  metric_writer->SetFileName(filename);
  metric_writer->SetInput(ug);
  metric_writer->Write();
  metric_writer->Delete();
  
  ug->GetPointData()->RemoveArray("mean_desired_lengths");
  ug->GetPointData()->RemoveArray("desired_lengths");

  int total_adapt_iterations = 4;
  double sum_time = 0;
  for(int adapt_iter=0;adapt_iter<total_adapt_iterations;adapt_iter++){

    Adaptivity adapt;
    adapt.SetInput(ug, NPrivateNodes[rank]);
    adapt.set_adapt_sweeps(10);
    adapt.set_surface_mesh(SENList);
    adapt.set_surface_ids(sids);
    double tic = MPI::Wtime();
    adapt.adapt();
    double toc = MPI::Wtime();
    sum_time += (toc-tic);
    if(rank==0)
      cerr<<"Adapt: "<<adapt_iter<<" time: "<<toc-tic<<endl;

    adapt.get_surface_ids(sids);
    adapt.get_surface_mesh(SENList);
    vtkUnstructuredGrid *new_ug = adapt.get_adapted_vtu(NPrivateNodes[rank]);
    ug = new_ug;

    vtkXMLUnstructuredGridWriter *ug_writer = vtkXMLUnstructuredGridWriter::New();
    sprintf(filename, "adapted_%d_%d.vtu", rank, adapt_iter+1);
    ug_writer->SetFileName(filename);
    ug_writer->SetInput(ug);
    ug_writer->Write();
    
    ug_writer->Delete();
    
    // Data import.
    sam_init(ug, NPrivateNodes[rank], SENList, sids);
    
    // Load-balancing and data migration.
    tic = MPI::Wtime();
    sam_migrate(adapt_iter<(total_adapt_iterations-1));
    toc = MPI::Wtime();
    sum_time += (toc-tic);
    if(rank==0)
      cerr<<"Migrate: "<<adapt_iter<<" time: "<<toc-tic<<endl;
    ug = sam_export_mesh(NPrivateNodes[rank], SENList, sids);
    
    vtkXMLPUnstructuredGridWriter *writer=vtkXMLPUnstructuredGridWriter::New();
    sprintf(filename, "load_balanced_%d.pvtu", adapt_iter+1);
    writer->SetFileName(filename);
    writer->SetDataModeToBinary();
    writer->SetNumberOfPieces(NProcs);
    writer->SetGhostLevel(1);
    writer->SetStartPiece(rank);
    writer->SetEndPiece(rank);
    writer->SetInput(ug);
    
    vtkZLibDataCompressor* compressor = vtkZLibDataCompressor::New();
    writer->SetCompressor(compressor);
    
    writer->Write();
    writer->Delete();
    compressor->Delete();
    
    // Cleanup
    sam_cleanup();
  }
  cerr<<"Total time: "<<sum_time<<endl;

  ug_reader->Delete();

  MPI::Finalize();
  return 0;
}
